/*
 * src/reasoning/indefinite/IndefinitePLNFormulas.h
 *
 * Modified by Cesar Marcondes <cesar@cs.ucla.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License v3 as
 * published by the Free Software Foundation and including the exceptions
 * at http://opencog.org/wiki/Licenses
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, write to:
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _INDEFINITEPLNFORMULAS_H
#define _INDEFINITEPLNFORMULAS_H

#include <IndefiniteTruthValue.h>
#include <vector>

using namespace std;
using namespace opencog;

// ***************************************************
// Syntactic sugar

typedef float number_t;
typedef vector<number_t> dvector;
typedef vector<number_t*> pvector;
typedef vector<vector<number_t> > dmatrix;

namespace reasoning {

// ***************************************************
// Decouple Sampling and Constraints From Core Indefinite Rules
class Sampler {
	public:
		Sampler(void *(*function)(void *), int n1_, int n2_) {
			void *(*distribution)(void *) = function;
			n1 = n1_;
			n2 = n2_;
			deduction = false;
		};

		void generateSample(IndefiniteTruthValue* const& TVa, 
				IndefiniteTruthValue* const& TVb);

		void generateSample(IndefiniteTruthValue* const& TVa, 
				IndefiniteTruthValue* const& TVb, 
				IndefiniteTruthValue* const& TVc);

		void generateSample(IndefiniteTruthValue* const& TVa, 
				IndefiniteTruthValue* const& TVb, 
				IndefiniteTruthValue* const& TVc,
				IndefiniteTruthValue* const& TVab, 
				IndefiniteTruthValue* const& TVbc);
		
		void setDimensions(int n1_, int n2_) { 
			n1 = n1_;
			n2 = n2_;
		};

		void setDeduction(bool b) {
			deduction = b;
		}
		
	private:
		bool deduction;
		int n1, n2;
		void *(*distribution)(void *arg);
};

// IndefiniteRule Super-class
class IndefiniteRule {
	public:
		IndefiniteRule ();
		
		IndefiniteRule (IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVb);

		IndefiniteRule (IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVb, 
						IndefiniteTruthValue* const& TVc);

		IndefiniteRule (IndefiniteTruthValue* const& TVa,
					    IndefiniteTruthValue* const& TVb, 
						IndefiniteTruthValue* const& TVc,
						IndefiniteTruthValue* const& TVab, 
						IndefiniteTruthValue* const& TVbc,
						bool deduction);

		virtual IndefiniteTruthValue* solve();
		virtual IndefiniteTruthValue* conclusion(dmatrix d);

	protected:
		IndefiniteTruthValue **tvset;
		Sampler *s;

};

// ***************************************************
// Sub-classes: ConjunctionRule, ImplicationRule, Revision
// 				AdbductionRule, BayesRule, DeductionRule

class ConjunctionRule : public IndefiniteRule {
	public:
		ConjunctionRule(IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVb);
		IndefiniteTruthValue* solve();
};

class ImplicationRule : public IndefiniteRule {
	public:
		ImplicationRule(IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVb);
		IndefiniteTruthValue* solve();
		IndefiniteTruthValue* conclusion(dmatrix d);
};

class RevisionRule : public IndefiniteRule {
	public:
		RevisionRule   (IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVb);
		IndefiniteTruthValue* solve();
};

class BayesRule : public IndefiniteRule {
	public:
		BayesRule	   (IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVc, 
						IndefiniteTruthValue* const& TVac);
		IndefiniteTruthValue* solve();
};

class AbductionRule : public IndefiniteRule {
	public:
		AbductionRule  (IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVb, 
						IndefiniteTruthValue* const& TVc, 
						IndefiniteTruthValue* const& TVab,
						IndefiniteTruthValue* const& TVbc);
		IndefiniteTruthValue* solve();
};

class DeductionRule : public IndefiniteRule {
	public:
		DeductionRule  (IndefiniteTruthValue* const& TVa, 
						IndefiniteTruthValue* const& TVc, 
						IndefiniteTruthValue* const& TVb, 
						IndefiniteTruthValue* const& TVab,
						IndefiniteTruthValue* const& TVbc);
		IndefiniteTruthValue* solve();
};

// ***************************************************
// Factory can create any IndefiniteRule sub-classing

class RuleGenerator {
	public:
		virtual IndefiniteRule* CreateRule() = 0;
};

template <typename R>
class IndefiniteRuleGenerator: public RuleGenerator {
	public:
		virtual IndefiniteRule* CreateRule();
};

template <typename R>
IndefiniteRule* IndefiniteRuleGenerator<R>::CreateRule() {
	return new R;
};

}; // namespace reasoning
#endif /*INDEFINITEPLNFORMULAS_H */
